#!/usr/bin/env python
import unittest
import sys
import os
from pathlib import Path
import json
import pandas as pd
import numpy as np

import emod_api.campaign as camp
import emodpy_tbhiv.demographics.TBHIVDemographics as Demographics

parent = Path(__file__).resolve().parent
sys.path.append(str(parent))
import manifest


class HIVDemographicsTest(unittest.TestCase):
    schema_path = manifest.schema_file

    # region unittest setup and teardown method
    @classmethod
    def setUpClass(cls):
        camp.set_schema( cls.schema_path )
        cls.out_folder = os.path.join(manifest.current_directory, 'output', 'demographics')

    def setUp(self):
        print(f"running {self._testMethodName}:")
        pass

    def tearDown(self):
        print("end of test\n")
        pass
    # endregion

    def check_for_unique_node_id(self, nodes):
        node_ids = list()
        for node in nodes:
            node_id = node['NodeID']
            if node_id not in node_ids:
                node_ids.append(node_id)
            else:
                return False
        return True

    def default_test(self, demog):
        defaults = demog.raw['Defaults']
        self.assertIn('NodeAttributes', defaults)
        self.assertIn('IndividualAttributes', defaults)
        ia = defaults['IndividualAttributes']
        self.assertIn('RiskDistributionFlag', ia)
        self.assertIn('RiskDistribution1', ia)
        self.assertIn('RiskDistribution2', ia)
        self.assertEqual(ia['RiskDistributionFlag'], 1)
        self.assertEqual(ia['RiskDistribution1'], 1)
        self.assertEqual(ia['RiskDistribution2'], 1)
        self.assertIn('AgeDistributionFlag', ia)
        self.assertIn('HIVCoinfectionDistribution', ia)
        self.assertIn('HIVTBCoinfMortalityDistribution', ia)

    # region unittests
    def test_from_basic_node(self):
        out_filename = os.path.join(self.out_folder, "demographics_basic_node.json")
        lat = 11
        lon = 22
        self.pop = 999
        name = "test_name"
        forced_id = 101
        demog = Demographics.fromBasicNode(lat=lat, lon=lon, pop=self.pop, name=name, forced_id=forced_id)
        self.assertTrue(isinstance(demog, Demographics.TBHIVDemographics))
        self.default_test(demog)

        print(f"Writing out file: {out_filename}.")
        demog.generate_file(out_filename)
        self.assertTrue(os.path.isfile(out_filename), msg=f'{out_filename} is not generated.')
        with open(out_filename, 'r') as demo_file:
            demog_json = json.load(demo_file)
        self.assertEqual(demog_json['Nodes'][0]['NodeAttributes']['Latitude'], lat)
        self.assertEqual(demog_json['Nodes'][0]['NodeAttributes']['Longitude'], lon)
        self.assertEqual(demog_json['Nodes'][0]['NodeAttributes']['InitialPopulation'], self.pop)
        self.assertEqual(demog_json['Nodes'][0]['NodeAttributes']['FacilityName'], name)
        self.assertEqual(demog_json['Nodes'][0]['NodeID'], forced_id)

    def test_from_data(self):  # infer_natural_mortality
        self.male_input_file = os.path.join(manifest.current_directory, "../data/Malawi_male_mortality.csv")
        self.female_input_file = os.path.join(manifest.current_directory, "../data/Malawi_female_mortality.csv")
        # These are hardcoded values
        interval_fit = [1970, 1980]
        self.predict_horizon = 2050
        self.results_scale_factor = 1.0 / 365.0

        self.pop = 1e6
        demog = Demographics.fromData(pop=self.pop, filename_male=self.male_input_file, filename_female=self.female_input_file)
        self.assertEqual(demog.nodes[0].pop, self.pop)
        self.check_everything( demog )

    def test_from_csv(self):
        self.init_pop_by_node_file = os.path.join(manifest.current_directory, "../data/init_pop_by_node.csv")
        self.male_input_file = os.path.join(manifest.current_directory, "../data/Malawi_male_mortality.csv")
        self.female_input_file = os.path.join(manifest.current_directory, "../data/Malawi_female_mortality.csv")
        # These are hardcoded values
        interval_fit = [1970, 1980]
        self.predict_horizon = 2050
        self.results_scale_factor = 1.0 / 365.0

        id_reference = 'from_csv' # hardcoded value
        demog = Demographics.from_csv(pop_file=self.init_pop_by_node_file, filename_male=self.male_input_file, filename_female=self.female_input_file, id_ref=id_reference)
        self.check_everything( demog )

        # The code below was grabbed from emodpy-malaria 2/22/22
        out_filename = os.path.join(self.out_folder, "demographics_from_csv.json")
        demog.generate_file(out_filename)
        with open(out_filename, 'r') as demo_file:
            demog_json = json.load(demo_file)

        # Checking we can grab a node
        inspect_node = demog.get_node(demog.nodes[0].id)
        self.assertEqual(inspect_node.id, demog.nodes[0].id, msg=f"This node should have an id of {demog.nodes[0].id} "
                                                                 f"but instead it is {inspect_node.id}")

        self.assertEqual(demog_json['Metadata']['IdReference'], id_reference)

        self.assertDictEqual(demog_json, demog.raw)

        import pandas as pd
        csv_df = pd.read_csv(self.init_pop_by_node_file, encoding='iso-8859-1')

        # checking if we have the same number of nodes and the number of rows in csv file
        self.assertEqual(len(csv_df), len(demog_json['Nodes']))

        self.assertTrue(self.check_for_unique_node_id(demog.raw['Nodes']))
        for index, row in csv_df.iterrows():
            pop = int(row['pop'])
            lat = float(row['lat'])
            lon = float(row['lon'])
            node_id = int(row['node_id'])
            self.assertEqual(pop, demog.nodes[index].node_attributes.initial_population)
            self.assertEqual(lat, demog.nodes[index].node_attributes.latitude)
            self.assertEqual(lon, demog.nodes[index].node_attributes.longitude)
            self.assertEqual(node_id, demog.nodes[index].forced_id)

    def test_from_params(self):
        out_filename = os.path.join(self.out_folder, "demographics_from_params.json")
        self.male_input_file = os.path.join(manifest.current_directory, "../data/Malawi_male_mortality.csv")
        self.female_input_file = os.path.join(manifest.current_directory, "../data/Malawi_female_mortality.csv")

        totpop = 1e5
        num_nodes = 250
        frac_rural = 0.1
        self.predict_horizon = 2050
        self.results_scale_factor = 1.0 / 365.0
        id_reference = 'from_params'
        demog = Demographics.from_params(tot_pop=totpop, num_nodes=num_nodes, frac_rural=frac_rural, filename_male=self.male_input_file, filename_female=self.female_input_file, id_ref=id_reference)
        demog.generate_file(out_filename)
        self.check_everything( demog )

        self.assertTrue(os.path.isfile(out_filename), msg=f'{out_filename} is not generated.')
        with open(out_filename, 'r') as demo_file:
            demog_json = json.load(demo_file)

        self.assertEqual(demog_json['Metadata']['IdReference'], id_reference)

        self.assertDictEqual(demog_json, demog.raw)

        self.assertEqual(num_nodes, len(demog_json['Nodes']))

        sum_pop = 0
        for node in demog_json['Nodes']:
            sum_pop += node['NodeAttributes']['InitialPopulation']
        # Todo: add this assertion back when #112 is fixed.
        # self.assertEqual(sum_pop, totpop)
        if sum_pop != totpop:
            print(f"Something went wrong, expected totpop is {totpop}, got {sum_pop} total population.")

        self.assertTrue(self.check_for_unique_node_id(demog.raw['Nodes']))

    def check_everything( self, demog ):
        self.assertTrue(isinstance(demog, Demographics.TBHIVDemographics))
        self.default_test(demog)

        male_input = pd.read_csv(self.male_input_file)
        female_input = pd.read_csv(self.female_input_file)
        output = demog.raw['Defaults']

        # Check pop is honored
        #self.assertEqual(demog.nodes[0].pop, self.pop)

        # Check population group consistency
        male_distribution = output['IndividualAttributes']['MortalityDistributionMale']
        female_distribution = output['IndividualAttributes']['MortalityDistributionFemale']

        # Check age group
        male_age_groups = male_distribution['NumPopulationGroups'][0]
        self.assertEqual(len(male_distribution['PopulationGroups'][0]), male_age_groups)
        # Get age groups information male from csv file
        expected_male_age_group = np.append(male_input['Age (x)'].unique(), 100).tolist()
        self.assertListEqual(expected_male_age_group[1:], male_distribution['PopulationGroups'][0])

        female_age_groups = female_distribution['NumPopulationGroups'][0]
        self.assertEqual(len(female_distribution['PopulationGroups'][0]), female_age_groups)
        # Get age groups information from female csv file
        expected_female_age_group = np.append(female_input['Age (x)'].unique(), 100).tolist()
        self.assertListEqual(expected_female_age_group[1:], female_distribution['PopulationGroups'][0])

        # Check Year
        male_year_groups = male_distribution['NumPopulationGroups'][1]
        self.assertEqual(len(male_distribution['PopulationGroups'][1]), male_year_groups)
        # Get year groups information male from csv file
        expected_male_year_group = male_input['Ave_Year'].unique().tolist()
        # Up to year = predict_horizon
        expected_male_year_group = [x for x in expected_male_year_group if x < self.predict_horizon]
        self.assertListEqual(expected_male_year_group, male_distribution['PopulationGroups'][1])

        female_year_groups = female_distribution['NumPopulationGroups'][1]
        self.assertEqual(len(female_distribution['PopulationGroups'][1]), female_year_groups)
        expected_female_year_group = expected_male_year_group
        self.assertListEqual(expected_female_year_group, female_distribution['PopulationGroups'][1])

        # Check prediction horizon is honored
        self.assertLessEqual(max(male_distribution['PopulationGroups'][1]), self.predict_horizon)
        self.assertLessEqual(max(female_distribution['PopulationGroups'][1]), self.predict_horizon)

        # Check results scale factor is consistent with parameters
        self.assertEqual(male_distribution['ResultScaleFactor'], self.results_scale_factor)
        self.assertEqual(female_distribution['ResultScaleFactor'], self.results_scale_factor)

        # Check result values array length
        self.assertEqual(len(male_distribution['ResultValues']), male_age_groups)
        for m_y in male_distribution['ResultValues']:
            self.assertEqual(len(m_y), male_year_groups)
        self.assertEqual(len(female_distribution['ResultValues']), female_age_groups)
        for f_y in female_distribution['ResultValues']:
            self.assertEqual(len(f_y), female_year_groups)

        # Check result values consistency with reference files
        male_reference = pd.read_csv(os.path.join(self.out_folder, "MaleTrue"))
        female_reference = pd.read_csv(os.path.join(self.out_folder, "FemaleTrue"))
        for i in range(male_age_groups):
            for j in range(male_year_groups):
                male_mortality_rate = male_distribution['ResultValues'][i][j]
                expected_male_mortality_rate = male_reference[male_reference['Age'] == expected_male_age_group[i + 1]][
                    str(expected_male_year_group[j])].iloc[0]
                self.assertAlmostEqual(expected_male_mortality_rate, male_mortality_rate, delta=1e-5,
                                       msg=f"at year {expected_male_year_group[j]} age {expected_male_age_group[i + 1]}"
                                           f", male mortality rate is set to {male_mortality_rate} (please see "
                                           f"male_distribution['ResultValues'][{i}][{j}]), while it's "
                                           f"{expected_male_mortality_rate} in male csv file.\n")

                female_mortality_rate = female_distribution['ResultValues'][i][j]
                expected_female_mortality_rate = female_reference[female_reference['Age'] ==
                                                                  expected_female_age_group[i + 1]][
                    str(expected_female_year_group[j])].iloc[0]
                self.assertAlmostEqual(expected_female_mortality_rate, female_mortality_rate, delta=1e-5,
                                       msg=f"at year {expected_female_year_group[j]} age "
                                           f"{expected_female_age_group[i + 1]},"
                                           f" female mortality rate is set to {female_mortality_rate} (please see "
                                           f"female_distribution['ResultValues'][{i}][{j}]), while it's "
                                           f"{expected_female_mortality_rate} in female csv file.\n")
    # endregion


if __name__ == '__main__':
    unittest.main()

